# Copyright (C) 2003-2019, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

import datetime
import ftplib
import stat
import time

import freezegun
import pytest

import ftputil
import ftputil.error
import ftputil.stat
from ftputil.stat import MINUTE_PRECISION, DAY_PRECISION, UNKNOWN_PRECISION

from test import test_base
from test import scripted_session


Call = scripted_session.Call


# Special value to handle special case of datetimes before the epoch.
EPOCH = time.gmtime(0)[:6]


def stat_tuple_to_seconds(t):
    """
    Return a float number representing the UTC timestamp from the six-element
    tuple `t`.
    """
    assert len(t) == 6, "need a six-element tuple (year, month, day, hour, min, sec)"
    # Do _not_ apply `time.mktime` to the `EPOCH` value below. On some
    # platforms (e. g. Windows) this might cause an `OverflowError`.
    if t == EPOCH:
        return 0.0
    else:
        return datetime.datetime(*t, tzinfo=datetime.timezone.utc).timestamp()


class TestParsers:

    #
    # Helper methods
    #
    def _test_valid_lines(self, parser_class, lines, expected_stat_results):
        parser = parser_class()
        for line, expected_stat_result in zip(lines, expected_stat_results):
            # Convert to list to compare with the list `expected_stat_results`.
            parse_result = parser.parse_line(line, time_shift=5 * 60 * 60)
            stat_result = list(parse_result) + [
                parse_result._st_mtime_precision,
                parse_result._st_name,
                parse_result._st_target,
            ]
            # Convert time tuple to seconds.
            expected_stat_result[8] = stat_tuple_to_seconds(expected_stat_result[8])
            # Compare lists.
            assert stat_result == expected_stat_result

    def _test_invalid_lines(self, parser_class, lines):
        parser = parser_class()
        for line in lines:
            with pytest.raises(ftputil.error.ParserError):
                parser.parse_line(line)

    def _expected_year(self):
        """
        Return the expected year for the second line in the listing in
        `test_valid_unix_lines`.
        """
        # If in this year it's after Dec 19, 23:11, use the current year, else
        # use the previous year. This datetime value corresponds to the
        # hard-coded value in the string lists below.
        client_datetime = datetime.datetime.utcnow().replace(
            tzinfo=datetime.timezone.utc
        )
        server_datetime_candidate = client_datetime.replace(
            month=12, day=19, hour=23, minute=11, second=0
        )
        if server_datetime_candidate > client_datetime:
            return server_datetime_candidate.year - 1
        else:
            return server_datetime_candidate.year

    #
    # Unix parser
    #
    def test_valid_unix_lines(self):
        lines = [
            "drwxr-sr-x   2 45854    200           512 May  4  2000 "
            "chemeng link -> chemeng target",
            # The year value for this line will change with the actual time.
            "-rw-r--r--   1 45854    200          4604 Dec 19 23:11 index.html",
            "drwxr-sr-x   2 45854    200           512 Jan 01  2000 os2",
            "----------   2 45854    200           512 May 29  2000 some_file",
            "lrwxrwxrwx   2 45854    200           512 May 29  2000 osup -> " "../os2",
        ]
        # Note that the time shift is also subtracted from the datetimes that
        # have only day precision, i. e. a year but no time.
        expected_stat_results = [
            [
                17901,
                None,
                None,
                2,
                "45854",
                "200",
                512,
                None,
                (2000, 5, 3, 19, 0, 0),
                None,
                DAY_PRECISION,
                "chemeng link",
                "chemeng target",
            ],
            [
                33188,
                None,
                None,
                1,
                "45854",
                "200",
                4604,
                None,
                (self._expected_year(), 12, 19, 18, 11, 0),
                None,
                MINUTE_PRECISION,
                "index.html",
                None,
            ],
            [
                17901,
                None,
                None,
                2,
                "45854",
                "200",
                512,
                None,
                (1999, 12, 31, 19, 0, 0),
                None,
                DAY_PRECISION,
                "os2",
                None,
            ],
            [
                32768,
                None,
                None,
                2,
                "45854",
                "200",
                512,
                None,
                (2000, 5, 28, 19, 0, 0),
                None,
                DAY_PRECISION,
                "some_file",
                None,
            ],
            [
                41471,
                None,
                None,
                2,
                "45854",
                "200",
                512,
                None,
                (2000, 5, 28, 19, 0, 0),
                None,
                DAY_PRECISION,
                "osup",
                "../os2",
            ],
        ]
        self._test_valid_lines(ftputil.stat.UnixParser, lines, expected_stat_results)

    def test_alternative_unix_format(self):
        # See http://ftputil.sschwarzer.net/trac/ticket/12 for a description
        # for the need for an alternative format.
        lines = [
            "drwxr-sr-x   2   200           512 May  4  2000 "
            "chemeng link -> chemeng target",
            # The year value for this line will change with the actual time.
            "-rw-r--r--   1   200          4604 Dec 19 23:11 index.html",
            "drwxr-sr-x   2   200           512 May 29  2000 os2",
            "lrwxrwxrwx   2   200           512 May 29  2000 osup -> ../os2",
        ]
        expected_stat_results = [
            [
                17901,
                None,
                None,
                2,
                None,
                "200",
                512,
                None,
                (2000, 5, 3, 19, 0, 0),
                None,
                DAY_PRECISION,
                "chemeng link",
                "chemeng target",
            ],
            [
                33188,
                None,
                None,
                1,
                None,
                "200",
                4604,
                None,
                (self._expected_year(), 12, 19, 18, 11, 0),
                None,
                MINUTE_PRECISION,
                "index.html",
                None,
            ],
            [
                17901,
                None,
                None,
                2,
                None,
                "200",
                512,
                None,
                (2000, 5, 28, 19, 0, 0),
                None,
                DAY_PRECISION,
                "os2",
                None,
            ],
            [
                41471,
                None,
                None,
                2,
                None,
                "200",
                512,
                None,
                (2000, 5, 28, 19, 0, 0),
                None,
                DAY_PRECISION,
                "osup",
                "../os2",
            ],
        ]
        self._test_valid_lines(ftputil.stat.UnixParser, lines, expected_stat_results)

    def test_pre_epoch_times_for_unix(self):
        # See http://ftputil.sschwarzer.net/trac/ticket/83 .
        # `mirrors.ibiblio.org` returns dates before the "epoch" that cause an
        # `OverflowError` in `mktime` on some platforms, e. g. Windows.
        lines = [
            "-rw-r--r--   1 45854    200          4604 May  4  1968 index.html",
            "-rw-r--r--   1 45854    200          4604 Dec 31  1969 index.html",
            "-rw-r--r--   1 45854    200          4604 May  4  1800 index.html",
        ]
        expected_stat_result = [
            33188,
            None,
            None,
            1,
            "45854",
            "200",
            4604,
            None,
            EPOCH,
            None,
            UNKNOWN_PRECISION,
            "index.html",
            None,
        ]
        # Make shallow copies to avoid converting the time tuple more than once
        # in _test_valid_lines`.
        expected_stat_results = [
            expected_stat_result[:],
            expected_stat_result[:],
            expected_stat_result[:],
        ]
        self._test_valid_lines(ftputil.stat.UnixParser, lines, expected_stat_results)

    def test_invalid_unix_lines(self):
        lines = [
            # Not intended to be parsed. Should have been filtered out by
            # `ignores_line`.
            "total 14",
            # Invalid month abbreviation
            "drwxr-sr-x   2 45854    200           512 Max  4  2000 chemeng",
            # Year value isn't an integer
            "drwxr-sr-x   2 45854    200           512 May  4  abcd chemeng",
            # Day value isn't an integer
            "drwxr-sr-x   2 45854    200           512 May ab  2000 chemeng",
            # Hour value isn't an integer
            "-rw-r--r--   1 45854    200          4604 Dec 19 ab:11 index.html",
            # Minute value isn't an integer
            "-rw-r--r--   1 45854    200          4604 Dec 19 23:ab index.html",
            # Day value too large
            "drwxr-sr-x   2 45854    200           512 May 32  2000 chemeng",
            # Ditto, for time instead of year
            "drwxr-sr-x   2 45854    200           512 May 32 11:22 chemeng",
            # Incomplete mode
            "drwxr-sr-    2 45854    200           512 May  4  2000 chemeng",
            # Invalid first letter in mode
            "xrwxr-sr-x   2 45854    200           512 May  4  2000 chemeng",
            # Ditto, plus invalid size value
            "xrwxr-sr-x   2 45854    200           51x May  4  2000 chemeng",
            # Is this `os1 -> os2` pointing to `os3`, or `os1` pointing to
            # `os2 -> os3` or the plain name `os1 -> os2 -> os3`? We don't
            # know, so we consider the line invalid.
            "drwxr-sr-x   2 45854    200           512 May 29  2000 "
            "os1 -> os2 -> os3",
            # Missing name
            "-rwxr-sr-x   2 45854    200           51x May  4  2000 ",
        ]
        self._test_invalid_lines(ftputil.stat.UnixParser, lines)

    #
    # Microsoft parser
    #
    def test_valid_ms_lines_two_digit_year(self):
        lines = [
            "07-27-01  11:16AM       <DIR>          Test",
            "10-23-95  03:25PM       <DIR>          WindowsXP",
            "07-17-00  02:08PM             12266720 test.exe",
            "07-17-09  12:08AM             12266720 test.exe",
            "07-17-09  12:08PM             12266720 test.exe",
        ]
        expected_stat_results = [
            [
                16640,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                (2001, 7, 27, 6, 16, 0),
                None,
                MINUTE_PRECISION,
                "Test",
                None,
            ],
            [
                16640,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                (1995, 10, 23, 10, 25, 0),
                None,
                MINUTE_PRECISION,
                "WindowsXP",
                None,
            ],
            [
                33024,
                None,
                None,
                None,
                None,
                None,
                12266720,
                None,
                (2000, 7, 17, 9, 8, 0),
                None,
                MINUTE_PRECISION,
                "test.exe",
                None,
            ],
            [
                33024,
                None,
                None,
                None,
                None,
                None,
                12266720,
                None,
                (2009, 7, 16, 19, 8, 0),
                None,
                MINUTE_PRECISION,
                "test.exe",
                None,
            ],
            [
                33024,
                None,
                None,
                None,
                None,
                None,
                12266720,
                None,
                (2009, 7, 17, 7, 8, 0),
                None,
                MINUTE_PRECISION,
                "test.exe",
                None,
            ],
        ]
        self._test_valid_lines(ftputil.stat.MSParser, lines, expected_stat_results)

    def test_valid_ms_lines_four_digit_year(self):
        # See http://ftputil.sschwarzer.net/trac/ticket/67
        lines = [
            "10-19-2012  03:13PM       <DIR>          SYNCDEST",
            "10-19-2012  03:13PM       <DIR>          SYNCSOURCE",
            "10-19-1968  03:13PM       <DIR>          SYNC",
        ]
        expected_stat_results = [
            [
                16640,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                (2012, 10, 19, 10, 13, 0),
                None,
                MINUTE_PRECISION,
                "SYNCDEST",
                None,
            ],
            [
                16640,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                (2012, 10, 19, 10, 13, 0),
                None,
                MINUTE_PRECISION,
                "SYNCSOURCE",
                None,
            ],
            [
                16640,
                None,
                None,
                None,
                None,
                None,
                None,
                None,
                EPOCH,
                None,
                UNKNOWN_PRECISION,
                "SYNC",
                None,
            ],
        ]
        self._test_valid_lines(ftputil.stat.MSParser, lines, expected_stat_results)

    def test_invalid_ms_lines(self):
        lines = [
            # Neither "<DIR>" nor a size present
            "07-27-01  11:16AM                      Test",
            # "AM"/"PM" missing
            "07-17-00  02:08             12266720 test.exe",
            # Year not an int
            "07-17-ab  02:08AM           12266720 test.exe",
            # Month not an int
            "ab-17-00  02:08AM           12266720 test.exe",
            # Day not an int
            "07-ab-00  02:08AM           12266720 test.exe",
            # Hour not an int
            "07-17-00  ab:08AM           12266720 test.exe",
            # Invalid size value
            "07-17-00  02:08AM           1226672x test.exe",
        ]
        self._test_invalid_lines(ftputil.stat.MSParser, lines)

    #
    # The following code checks if the decision logic in the Unix line parser
    # for determining the year works.
    #
    def dir_line(self, datetime_):
        """
        Return a directory line as from a Unix FTP server. Most of the contents
        are fixed, but the timestamp is made from `time_float` (seconds since
        the epoch, as from `time.time()`).
        """
        line_template = "-rw-r--r--   1   45854   200   4604   {}   index.html"
        datetime_string = datetime_.strftime("%b %d %H:%M")
        return line_template.format(datetime_string)

    def assert_equal_times(self, time1, time2):
        """
        Check if both times (seconds since the epoch) are equal. For the
        purpose of this test, two times are "equal" if they differ no more than
        one minute from each other.
        """
        abs_difference = abs(time1 - time2)
        assert abs_difference <= 60.0, "Difference is %s seconds" % abs_difference

    def _test_time_shift(self, supposed_time_shift, deviation=0.0):
        """
        Check if the stat parser considers the time shift value correctly.
        `deviation` is the difference between the actual time shift and the
        supposed time shift, which is rounded to full hours.
        """
        script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            # Explicitly use Unix format parser here.
            host._stat._parser = ftputil.stat.UnixParser()
            host.set_time_shift(supposed_time_shift)
            server_time = datetime.datetime.utcnow().replace(
                tzinfo=datetime.timezone.utc
            ) + datetime.timedelta(seconds=supposed_time_shift + deviation)
            stat_result = host._stat._parser.parse_line(
                self.dir_line(server_time), host.time_shift()
            )
            # We expect `st_mtime` in UTC.
            self.assert_equal_times(
                stat_result.st_mtime,
                (
                    server_time
                    # Convert back to client time.
                    - datetime.timedelta(seconds=supposed_time_shift)
                ).timestamp(),
            )

    def test_time_shifts(self):
        """
        Test correct year depending on time shift value.
        """
        # 1. test: Client and server share the same time (UTC). This is true if
        # the directory listing from the server is in UTC.
        self._test_time_shift(0.0)
        # 2. test: Server is three hours ahead of client
        self._test_time_shift(3 * 60 * 60)
        # Ditto, but with client and server in different years. See ticket #131.
        with freezegun.freeze_time("2019-12-31 22:37"):
            self._test_time_shift(3 * 60 * 60)
        # 3. test: Client is three hours ahead of server
        self._test_time_shift(-3 * 60 * 60)
        # 4. test: Server is supposed to be three hours ahead, but is ahead
        # three hours and one minute
        self._test_time_shift(3 * 60 * 60, 60)
        # 5. test: Server is supposed to be three hours ahead, but is ahead
        # three hours minus one minute
        self._test_time_shift(3 * 60 * 60, -60)
        # 6. test: Client is supposed to be three hours ahead, but is ahead
        # three hours and one minute
        self._test_time_shift(-3 * 60 * 60, -60)
        # 7. test: Client is supposed to be three hours ahead, but is ahead
        # three hours minus one minute
        self._test_time_shift(-3 * 60 * 60, 60)


class TestLstatAndStat:
    """
    Test `FTPHost.lstat` and `FTPHost.stat` (test currently only implemented
    for Unix server format).
    """

    def test_repr(self):
        """
        Test if the `repr` result looks like a named tuple.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result="drwxr-sr-x   2 45854   200   512 May  4  2000 foo",
            ),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            stat_result = host.stat("/foo")
            expected_result = (
                "StatResult(st_mode=17901, st_ino=None, st_dev=None, "
                "st_nlink=2, st_uid='45854', st_gid='200', st_size=512, "
                "st_atime=None, st_mtime=957398400.0, st_ctime=None)"
            )
            assert repr(stat_result) == expected_result

    def test_failing_lstat(self):
        """
        Test whether `lstat` fails for a nonexistent path.
        """
        # Directory with presumed file item doesn't exist.
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=""),
            Call("cwd", args=("/",)),
            # See FIXME comment in `ftputil.stat._Stat._real_lstat`
            Call("cwd", args=("/",)),
            Call("cwd", args=("/notthere",), result=ftplib.error_perm),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            with pytest.raises(ftputil.error.PermanentError):
                host.lstat("/notthere/irrelevant")
        # Directory exists, but not the file system item in the directory.
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result=test_base.dir_line(
                    mode_string="dr-xr-xr-x",
                    datetime_=datetime.datetime.now(),
                    name="some_dir",
                ),
            ),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/some_dir",)),
            Call("dir", args=("",), result=""),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            with pytest.raises(ftputil.error.PermanentError):
                host.lstat("/some_dir/notthere")

    def test_lstat_for_root(self):
        """
        Test `lstat` for `/` .

        Note: `(l)stat` works by going one directory up and parsing the output
        of an FTP `LIST` command. Unfortunately, it's not possible to do this
        for the root directory `/`.
        """
        script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            with pytest.raises(ftputil.error.RootDirError) as exc_info:
                host.lstat("/")
        # `RootDirError` is "outside" the `FTPOSError` hierarchy.
        assert not isinstance(exc_info.value, ftputil.error.FTPOSError)
        del exc_info

    def test_lstat_one_unix_file(self):
        """
        Test `lstat` for a file described in Unix-style format.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result="-rw-r--r--   1 45854   200   4604 Jan 19 23:11 some_file",
            ),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            stat_result = host.lstat("/some_file")
        assert oct(stat_result.st_mode) == "0o100644"
        assert stat_result.st_size == 4604
        assert stat_result._st_mtime_precision == 60

    def test_lstat_one_ms_file(self):
        """
        Test `lstat` for a file described in DOS-style format.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            # First try with unix parser, but this parser can't parse this
            # line.
            Call(
                "dir",
                args=("",),
                result="07-17-00  02:08PM             12266720 some_file",
            ),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            # Now try with MS parser.
            Call(
                "dir",
                args=("",),
                result="07-17-00  02:08PM             12266720 some_file",
            ),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            stat_result = host.lstat("/some_file")
        assert stat_result._st_name == "some_file"
        assert stat_result._st_mtime_precision == 60

    def test_lstat_one_unix_dir(self):
        """
        Test `lstat` for a directory described in Unix-style format.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result="drwxr-sr-x   6 45854   200   512 Sep 20  1999 some_dir",
            ),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            stat_result = host.lstat("/some_dir")
        assert oct(stat_result.st_mode) == "0o42755"
        assert stat_result.st_ino is None
        assert stat_result.st_dev is None
        assert stat_result.st_nlink == 6
        assert stat_result.st_uid == "45854"
        assert stat_result.st_gid == "200"
        assert stat_result.st_size == 512
        assert stat_result.st_atime is None
        assert stat_result.st_mtime == stat_tuple_to_seconds((1999, 9, 20, 0, 0, 0))
        assert stat_result.st_ctime is None
        assert stat_result._st_mtime_precision == 24 * 60 * 60
        assert stat_result == (
            17901,
            None,
            None,
            6,
            "45854",
            "200",
            512,
            None,
            stat_tuple_to_seconds((1999, 9, 20, 0, 0, 0)),
            None,
        )

    def test_lstat_one_ms_dir(self):
        """
        Test `lstat` for a directory described in DOS-style format.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            # First try with unix parser, but this parser can't parse this
            # line.
            Call(
                "dir",
                args=("",),
                result="10-23-01  03:25PM       <DIR>          some_dir",
            ),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            # Now try with MS parser.
            Call(
                "dir",
                args=("",),
                result="10-23-01  03:25PM       <DIR>          some_dir",
            ),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            stat_result = host.lstat("/some_dir")
        assert stat_result._st_mtime_precision == 60

    def test_lstat_via_stat_module(self):
        """
        Test `lstat` indirectly via `stat` module.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result="drwxr-sr-x   2 45854   200   512 May  4  2000 some_dir",
            ),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            stat_result = host.lstat("/some_dir")
        assert stat.S_ISDIR(stat_result.st_mode)

    def test_stat_following_link(self):
        """
        Test `stat` when invoked on a link.
        """
        # Simple link
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result="lrwxrwxrwx   1 45854   200   21 Jan 19  2002 link -> link_target",
            ),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result="-rw-r--r--   1 45854   200   4604 Jan 19 23:11 link_target",
            ),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            stat_result = host.stat("/link")
        assert stat_result.st_size == 4604
        # Link pointing to a link
        dir_lines = (
            "lrwxrwxrwx   1 45854   200   7    Jan 19  2002 link_link -> link\n"
            "lrwxrwxrwx   1 45854   200   14   Jan 19  2002 link -> link_target\n"
            "-rw-r--r--   1 45854   200   4604 Jan 19 23:11 link_target"
        )
        # Note that only one `dir` call would be needed in case of an enabled
        # cache.
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            # Look up `/link_link`.
            Call("dir", args=("",), result=dir_lines),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            # Look up `/link`.
            Call("dir", args=("",), result=dir_lines),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            # Look up `/link_target`.
            Call("dir", args=("",), result=dir_lines),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            stat_result = host.stat("/link_link")
        assert stat_result.st_size == 4604
        # Recursive link structure
        dir_lines = (
            "lrwxrwxrwx   1 45854   200   7    Jan 19  2002 bad_link1 -> bad_link2\n"
            "lrwxrwxrwx   1 45854   200   14   Jan 19  2002 bad_link2 -> bad_link1"
        )
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            # This dir finds the `bad_link1` name requested in the `stat` call.
            Call("dir", args=("",), result=dir_lines),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            # Look up link target `bad_link2`.
            Call("dir", args=("",), result=dir_lines),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            # FIXME: `stat` looks up the link target pointed to by `bad_link2`,
            # which is `bad_link1`. Only here ftputil notices the recursive
            # link chain. Obviously the start of the link chain hadn't been
            # stored in `visited_paths` (see also ticket #108).
            Call("dir", args=("",), result=dir_lines),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            with pytest.raises(ftputil.error.PermanentError):
                host.stat("bad_link1")

    #
    # Test automatic switching of Unix/MS parsers
    #
    def test_parser_switching_with_permanent_error(self):
        """
        Test non-switching of parser format with `PermanentError`.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir", args=("",), result="10-23-01  03:25PM       <DIR>          home"
            ),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir", args=("",), result="10-23-01  03:25PM       <DIR>          home"
            ),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            assert host._stat._allow_parser_switching is True
            # With these directory contents, we get a `ParserError` for the
            # Unix parser first, so `_allow_parser_switching` can be switched
            # off no matter whether we got a `PermanentError` afterward or not.
            with pytest.raises(ftputil.error.PermanentError):
                host.lstat("/nonexistent")
            assert host._stat._allow_parser_switching is False

    def test_parser_switching_default_to_unix(self):
        """
        Test non-switching of parser format; stay with Unix.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result="-rw-r--r--   1 45854   200   4604 Jan 19 23:11 some_file",
            ),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            assert host._stat._allow_parser_switching is True
            assert host._stat._allow_parser_switching is True
            assert isinstance(host._stat._parser, ftputil.stat.UnixParser)
            stat_result = host.lstat("some_file")
            # The Unix parser worked, so keep it.
            assert isinstance(host._stat._parser, ftputil.stat.UnixParser)
            assert host._stat._allow_parser_switching is False

    def test_parser_switching_to_ms(self):
        """
        Test switching of parser from Unix to MS format.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result="07-17-00  02:08PM             12266720 some_file",
            ),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result="07-17-00  02:08PM             12266720 some_file",
            ),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            assert host._stat._allow_parser_switching is True
            assert isinstance(host._stat._parser, ftputil.stat.UnixParser)
            # Parsing the directory `/` with the Unix parser fails, so switch
            # to the MS parser.
            stat_result = host.lstat("/some_file")
            assert isinstance(host._stat._parser, ftputil.stat.MSParser)
            assert host._stat._allow_parser_switching is False
            assert stat_result._st_name == "some_file"
            assert stat_result.st_size == 12266720

    def test_parser_switching_regarding_empty_dir(self):
        """Test switching of parser if a directory is empty."""
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=""),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            assert host._stat._allow_parser_switching is True
            # When the directory we're looking into doesn't give us any lines
            # we can't decide whether the first parser worked, because it
            # wasn't applied. So keep the parser for now.
            result = host.listdir("/")
            assert result == []
            assert host._stat._allow_parser_switching is True
            assert isinstance(host._stat._parser, ftputil.stat.UnixParser)


class TestListdir:
    """
    Test `FTPHost.listdir`.
    """

    def test_failing_listdir(self):
        """
        Test failing `FTPHost.listdir`.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=""),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            with pytest.raises(ftputil.error.PermanentError):
                host.listdir("notthere")

    def test_succeeding_listdir(self):
        """
        Test succeeding `FTPHost.listdir`.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result="drwxr-sr-x   2 45854   200    512 Jan  3 17:17 download\n"
                "drwxr-sr-x   2 45854   200    512 Jul 30 17:14 dir with spaces\n"
                "lrwxrwxrwx   2 45854   200      6 May 29  2000 link -> ../link_target\n"
                "-rw-r--r--   1 45854   200   4604 Jan 19 23:11 index.html",
            ),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            remote_file_list = host.listdir(".")
            # Do we have all expected "files"?
            assert len(remote_file_list) == 4
            expected_names = ["download", "dir with spaces", "link", "index.html"]
            for name in expected_names:
                assert name in remote_file_list
            assert len(host.stat_cache) == 0



# Copyright (C) 2014-2021, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

"""
Session factory factory (the two "factory" are intentional :-) ) for ftputil.
"""

import ftplib

import ftputil.tool


__all__ = ["session_factory"]


# In a way, it would be appropriate to call this function
# `session_factory_factory`, but that's cumbersome to use. Think of the
# function returning a session factory and the shorter name should be fine.
def session_factory(
    base_class=ftplib.FTP,
    port=21,
    use_passive_mode=None,
    *,
    encrypt_data_channel=True,
    encoding=None,
    debug_level=None,
):
    """
    Create and return a session factory according to the keyword arguments.

    base_class: Base class to use for the session class (e. g. `ftplib.FTP_TLS`
    or `M2Crypto.ftpslib.FTP_TLS`, the default is `ftplib.FTP`).

    port: Port number (integer) for the command channel (default 21). If you
    don't know what "command channel" means, use the default or use what the
    provider gave you as "the FTP port".

    use_passive_mode: If `True`, explicitly use passive mode. If `False`,
    explicitly don't use passive mode. If `None` (default), let the
    `base_class` decide whether it wants to use active or passive mode.

    encrypt_data_channel: If `True` (the default), call the `prot_p` method of
    the base class if it has the method. If `False` or `None` (`None` is the
    default), don't call the method.

    encoding: Encoding (str) to use for directory and file paths, or `None`.
    Unicode (`str`) paths will be encoded with this encoding. Bytes paths are
    assumed to be in this encoding. The default (equivalent to passing `None`)
    is to use the default encoding of the `base_class` argument. Note that this
    encoding has changed from Python 3.8 to 3.9.

    In Python 3.8 and lower, the default path encoding is "latin-1"; in Python
    3.9, the default path encoding is "utf-8". Therefore, if you want an
    encoding that's independent of the Python version, pass an explicit
    `encoding`.

    Using a non-`None` `encoding` is only supported if `base_class` is
    `ftplib.FTP` or a subclass of it.

    debug_level: Debug level (integer) to be set on a session instance. The
    default is `None`, meaning no debugging output.

    This function should work for the base classes `ftplib.FTP`,
    `ftplib.FTP_TLS`. Other base classes should work if they use the same API
    as `ftplib.FTP`.

    Usage example:

      my_session_factory = session_factory(
                             base_class=ftplib.FTP_TLS,
                             use_passive_mode=True,
                             encrypt_data_channel=True)
      with ftputil.FTPHost(host, user, password,
                           session_factory=my_session_factory) as host:
        ...
    """
    if not isinstance(base_class, type):
        raise ValueError(f"`base_class` must be a class, but is {base_class!r}")
    if (encoding is not None) and (not issubclass(base_class, ftplib.FTP)):
        raise ValueError(
            f"`encoding` is only supported for `ftplib.FTP` and subclasses, "
            f"but base class is {base_class!r}"
        )

    class Session(base_class):
        """
        Session factory class created by `session_factory`.
        """

        # In Python 3.8 and below, the `encoding` class attribute was never
        # documented, but setting it is the only way to set a custom encoding
        # for remote file system paths. Since we set the encoding on the class
        # level, all instances created from this class will share this
        # encoding. That's ok because the user asked for a specific encoding of
        # the _factory_ when calling `session_factory`.
        #
        # Python 3.9 is the first Python version to have a documented way to
        # set a custom encoding (per instance).
        #
        # XXX: The following heuristic doesn't cover the case that we run under
        # Python 3.8 or earlier _and_ have a base class with an `encoding`
        # argument. Also, the heuristic will fail if we run under Python 3.9,
        # but have a base class that overrides the constructor so that it
        # doesn't support the `encoding` argument anymore.
        def __init__(self, host, user, password):
            if (
                encoding is not None
            ) and ftputil.path_encoding.RUNNING_UNDER_PY39_AND_UP:
                super().__init__(encoding=encoding)
            else:
                super().__init__()
            self.connect(host, port)
            if debug_level is not None:
                self.set_debuglevel(debug_level)
            self.login(user, password)
            # `set_pasv` can be called with `True` (causing passive mode) or
            # `False` (causing active mode).
            if use_passive_mode is not None:
                self.set_pasv(use_passive_mode)
            if encrypt_data_channel and hasattr(base_class, "prot_p"):
                self.prot_p()

    if (encoding is not None) and not ftputil.path_encoding.RUNNING_UNDER_PY39_AND_UP:
        Session.encoding = encoding
    return Session



# Copyright (C) 2003-2021, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

"""
ftputil.error - exception classes and wrappers
"""

# pylint: disable=too-many-ancestors

import ftplib

import ftputil.path_encoding
import ftputil.tool
import ftputil.version


# You _can_ import these with `from ftputil.error import *`, - but it's _not_
# recommended.
__all__ = [
    "CommandNotImplementedError",
    "FTPIOError",
    "FTPOSError",
    "InaccessibleLoginDirError",
    "InternalError",
    "KeepAliveError",
    "NoEncodingError",
    "ParserError",
    "PermanentError",
    "RootDirError",
    "SyncError",
    "TemporaryError",
    "TimeShiftError",
]


class FTPError(Exception):
    """
    General ftputil error class.
    """

    def __init__(self, *args, original_error=None):
        super().__init__(*args)
        # `strerror`
        self.strerror = ""
        if original_error is not None:
            try:
                self.strerror = str(original_error)
            except Exception:
                # Consume all errors. If the `str` call fails, it's more
                # appropriate to ignore `original_error` than to raise an
                # exception while instantiating `FTPError`.
                pass
        elif args:
            # Assume the first argument is a string. It may be a byte string
            # though.
            try:
                self.strerror = ftputil.tool.as_str(
                    args[0], ftputil.path_encoding.DEFAULT_ENCODING
                )
            except TypeError:
                # `args[0]` isn't `str` or `bytes`.
                pass
        # `errno`
        self.errno = None
        try:
            self.errno = int(self.strerror[:3])
        except ValueError:
            # `int()` argument couldn't be converted to an integer.
            pass
        # `file_name`
        self.file_name = None

    def __str__(self):
        return "{}\nDebugging info: {}".format(
            self.strerror, ftputil.version.version_info
        )


# Internal errors are those that have more to do with the inner workings of
# ftputil than with errors on the server side.
class InternalError(FTPError):
    """Internal error."""

    pass


class RootDirError(InternalError):
    """Raised for generic stat calls on the remote root directory."""

    pass


class InaccessibleLoginDirError(InternalError):
    """May be raised if the login directory isn't accessible."""

    pass


class TimeShiftError(InternalError):
    """Raised for invalid time shift values."""

    pass


class ParserError(InternalError):
    """Raised if a line of a remote directory can't be parsed."""

    pass


class CacheMissError(InternalError):
    """Raised if a path isn't found in the cache."""

    pass


class NoEncodingError(InternalError):
    """Raised if session instances don't specify an encoding."""

    pass


# Currently not used
class KeepAliveError(InternalError):
    """Raised if the keep-alive feature failed."""

    pass


class FTPOSError(FTPError, OSError):
    """Generic FTP error related to `OSError`."""

    pass


class TemporaryError(FTPOSError):
    """Raised for temporary FTP errors (4xx)."""

    pass


class PermanentError(FTPOSError):
    """Raised for permanent FTP errors (5xx)."""

    pass


class CommandNotImplementedError(PermanentError):
    """Raised if the server doesn't implement a certain feature (502)."""

    pass


class RecursiveLinksError(PermanentError):
    """Raised if an infinite link structure is detected."""

    pass


# Currently not used
class SyncError(PermanentError):
    """Raised for problems specific to syncing directories."""

    pass


class FtplibErrorToFTPOSError:
    """
    Context manager to convert `ftplib` exceptions to exceptions derived from
    `FTPOSError`.
    """

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            # No exception
            return
        if isinstance(exc_value, ftplib.error_temp):
            raise TemporaryError(
                *exc_value.args, original_error=exc_value
            ) from exc_value
        elif isinstance(exc_value, ftplib.error_perm):
            # If `exc_value.args[0]` is present, assume it's a byte or unicode
            # string.
            if exc_value.args and ftputil.tool.as_str(
                exc_value.args[0], ftputil.path_encoding.DEFAULT_ENCODING
            ).startswith("502"):
                raise CommandNotImplementedError(
                    *exc_value.args, original_error=exc_value
                ) from exc_value
            else:
                raise PermanentError(
                    *exc_value.args, original_error=exc_value
                ) from exc_value
        elif isinstance(exc_value, ftplib.all_errors):
            raise FTPOSError(*exc_value.args, original_error=exc_value) from exc_value
        else:
            raise


ftplib_error_to_ftp_os_error = FtplibErrorToFTPOSError()


class FTPIOError(FTPError, IOError):
    """Generic FTP error related to `IOError`."""

    pass


class FtplibErrorToFTPIOError:
    """
    Context manager to convert `ftplib` exceptions to `FTPIOError` exceptions.
    """

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            # No exception
            return
        if isinstance(exc_value, ftplib.all_errors):
            raise FTPIOError(*exc_value.args, original_error=exc_value) from exc_value
        else:
            raise


ftplib_error_to_ftp_io_error = FtplibErrorToFTPIOError()



# Copyright (C) 2002-2021, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

import datetime
import errno
import ftplib
import io
import itertools
import os
import pickle
import posixpath
import random
import time
import unittest
import unittest.mock
import warnings

import pytest

import ftputil
import ftputil.error
import ftputil.file
import ftputil.path_encoding
import ftputil.tool
import ftputil.stat

from test import test_base
import test.scripted_session as scripted_session


Call = scripted_session.Call


#
# Helper function to generate random data
#
def binary_data():
    """
    Return a bytes object of length 10000, consisting of bytes from a pool of
    integer numbers in the range 0..255.
    """
    pool = list(range(0, 256))
    size = 10000
    integer_list = [random.choice(pool) for i in range(size)]
    return bytes(integer_list)


def as_bytes(string, encoding=ftputil.path_encoding.DEFAULT_ENCODING):
    return string.encode(encoding)


#
# Test cases
#

# For Python < 3.9, the `default_session_factory` is just `ftplib.FTP`. It's
# not worth it to test the factory then.
@pytest.mark.skipif(
    not ftputil.path_encoding.RUNNING_UNDER_PY39_AND_UP,
    reason="tests apply only to Python 3.9 and up",
)
class TestDefaultSessionFactory:
    def test_ftplib_FTP_subclass(self):
        """
        Test if the default factory is a subclass of `ftplib.FTP`.
        """
        assert issubclass(ftputil.host.default_session_factory, ftplib.FTP)

    def _test_extra_arguments(self, args=None, kwargs=None, expected_kwargs=None):
        """
        Test if `ftputil.FTPHost` accepts additional positional and keyword
        arguments, which are then passed to the session factory.
        """
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}
        if expected_kwargs is None:
            expected_kwargs = kwargs
        # Since our test server listens on a non-default port, we can't use the
        # session factory directly. We have to mock `ftplib.FTP` which is used
        # by ftputil's `default_session_factory`.
        with unittest.mock.patch("ftplib.FTP.__init__") as ftp_mock:
            # Prevent `TypeError` when Python checks the `__init__` result.
            ftp_mock.return_value = None
            session = ftputil.host.default_session_factory(
                "localhost", "ftptest", "dummy", *args, **kwargs
            )
            assert len(ftp_mock.call_args_list) == 1
            assert (
                # Don't compare the `self` argument. It changes for every test
                # run.
                ftp_mock.call_args.args
                == ("localhost", "ftptest", "dummy") + args
            )
            assert ftp_mock.call_args.kwargs == expected_kwargs

    def test_extra_positional_arguments(self):
        """
        Test if extra positional arguments are passed to the `ftplib.FTP`
        constructor.
        """
        expected_kwargs = {"encoding": ftputil.path_encoding.DEFAULT_ENCODING}
        # `acct`, `timeout`
        self._test_extra_arguments(
            args=("", 1.0), kwargs={}, expected_kwargs=expected_kwargs
        )

    def test_extra_keyword_arguments(self):
        """
        Test if extra keyword arguments are passed to the `ftplib.FTP`
        constructor.
        """
        kwargs = {"timeout": 1.0, "source_address": None}
        expected_kwargs = kwargs.copy()
        expected_kwargs["encoding"] = ftputil.path_encoding.DEFAULT_ENCODING
        self._test_extra_arguments(kwargs=kwargs, expected_kwargs=expected_kwargs)

    def test_custom_encoding(self):
        """
        Test if a custom encoding is passed to the base class constructor when
        running under Python 3.9 and up.
        """
        kwargs = {"timeout": 1.0, "source_address": None, "encoding": "latin-2"}
        self._test_extra_arguments(kwargs=kwargs)


class TestConstructor:
    """
    Test initialization of `FTPHost` objects.
    """

    def test_open_and_close(self):
        """
        Test if opening and closing an `FTPHost` object works as expected.
        """
        script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        host = test_base.ftp_host_factory(scripted_session.factory(script))
        host.close()
        assert host.closed is True
        assert host._children == []

    def test_invalid_login(self):
        """
        Login to invalid host must fail.
        """
        script = [Call("__init__", result=ftplib.error_perm), Call("pwd", result="/")]
        with pytest.raises(ftputil.error.FTPOSError):
            test_base.ftp_host_factory(scripted_session.factory(script))

    def test_pwd_normalization(self):
        """
        Test if the stored current directory is normalized.
        """
        script = [
            Call("__init__"),
            # Deliberately return the current working directory with a trailing
            # slash to test if it's removed when stored in the `FTPHost`
            # instance.
            Call("pwd", result="/home/"),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            assert host.getcwd() == "/home"

    def test_missing_encoding_attribute(self):
        """
        Test if a missing `encoding` attribute on the session instance leads to
        a `NoEncodingError`.
        """

        class InvalidSessionError:
            pass

        with pytest.raises(ftputil.error.NoEncodingError):
            _ = ftputil.host.FTPHost(session_factory=InvalidSessionError)


class TestKeepAlive:
    def test_succeeding_keep_alive(self):
        """
        Assume the connection is still alive.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            # `pwd` due to `keep_alive` call.
            Call("pwd", result="/"),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.keep_alive()

    def test_failing_keep_alive(self):
        """
        Assume the connection has timed out, so `keep_alive` fails.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/home"),
            # Simulate failing `pwd` call after the server closed the
            # connection due to a session timeout.
            Call("pwd", result=ftplib.error_temp),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            with pytest.raises(ftputil.error.TemporaryError):
                host.keep_alive()


class TestSetParser:
    class TrivialParser(ftputil.stat.Parser):
        """
        An instance of this parser always returns the same result from its
        `parse_line` method. This is all we need to check if ftputil uses the
        set parser. No actual parsing code is required here.
        """

        def __init__(self):
            # We can't use `os.stat("/home")` directly because we later need
            # the object's `_st_name` attribute, which we can't set on a
            # `os.stat` stat value.
            default_stat_result = ftputil.stat.StatResult(os.stat("/home"))
            default_stat_result._st_name = "home"
            self.default_stat_result = default_stat_result

        def parse_line(self, line, time_shift=0.0):
            return self.default_stat_result

    def test_set_parser(self):
        """
        Test if the selected parser is used.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                result="drwxr-xr-x   2 45854    200           512 May  4  2000 home",
            ),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            assert host._stat._allow_parser_switching is True
            trivial_parser = TestSetParser.TrivialParser()
            host.set_parser(trivial_parser)
            stat_result = host.stat("/home")
            assert stat_result == trivial_parser.default_stat_result
            assert host._stat._allow_parser_switching is False


class TestCommandNotImplementedError:
    def test_command_not_implemented_error(self):
        """
        Test if we get the anticipated exception if a command isn't implemented
        by the server.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            # `FTPHost.chmod` only raises a `CommandNotImplementedError` when
            # the exception text of the `ftplib.error_perm` starts with "502".
            Call(
                "voidcmd",
                result=ftplib.error_perm("502 command not implemented"),
                args=("SITE CHMOD 0644 nonexistent",),
            ),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            # `FTPHost.chmod` only raises a `CommandNotImplementedError` when
            # the exception text of the `ftplib.error_perm` starts with "502".
            Call(
                "voidcmd",
                result=ftplib.error_perm("502 command not implemented"),
                args=("SITE CHMOD 0644 nonexistent",),
            ),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            with pytest.raises(ftputil.error.CommandNotImplementedError):
                host.chmod("nonexistent", 0o644)
            # `CommandNotImplementedError` is a subclass of `PermanentError`.
            with pytest.raises(ftputil.error.PermanentError):
                host.chmod("nonexistent", 0o644)


class TestRecursiveListingForDotAsPath:
    """
    These tests are for issue #33, see
    http://ftputil.sschwarzer.net/trac/ticket/33 .
    """

    def test_plain_listing(self):
        """
        If an empty string is passed to `FTPHost._dir` it should be passed to
        `session.dir` unmodified.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=(".",)),
            # Check that the empty string is passed on to `session.dir`.
            Call("dir", args=("",), result="non-recursive listing"),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            lines = host._dir(host.curdir)
            assert lines[0] == "non-recursive listing"

    def test_empty_string_instead_of_dot_workaround(self):
        """
        If `FTPHost.listdir` is called with a dot as argument, the underlying
        `session.dir` should _not_ be called with the dot as argument, but with
        an empty string.
        """
        dir_result = (
            "total 10\n"
            "lrwxrwxrwx   1 staff     7 Aug 13  2003 bin -> usr/bin\n"
            "d--x--x--x   2 staff   512 Sep 24  2000 dev\n"
            "d--x--x--x   3 staff   512 Sep 25  2000 etc\n"
            "dr-xr-xr-x   3 staff   512 Oct  3  2000 pub\n"
            "d--x--x--x   5 staff   512 Oct  3  2000 usr\n"
        )
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_result),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            files = host.listdir(host.curdir)
            assert files == ["bin", "dev", "etc", "pub", "usr"]


class TestTimeShift:

    # Helper mock class that frees us from setting up complicated session
    # scripts for the remote calls.
    class _Path:
        def split(self, path):
            return posixpath.split(path)

        def set_mtime(self, mtime):
            self._mtime = mtime

        def getmtime(self, file_name):
            return self._mtime

        def join(self, *args):
            return posixpath.join(*args)

        def normpath(self, path):
            return posixpath.normpath(path)

        def isabs(self, path):
            return posixpath.isabs(path)

        def abspath(self, path):
            return "/_ftputil_sync_"

        # Needed for `isdir` in `FTPHost.remove`
        def isfile(self, path):
            return True

    def test_rounded_time_shift(self):
        """
        Test if time shift is rounded correctly.
        """
        script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        multisession_factory = scripted_session.factory(script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            # Use private bound method.
            rounded_time_shift = host._FTPHost__rounded_time_shift
            # Pairs consisting of original value and expected result
            test_data = [
                (0, 0),
                (0.1, 0),
                (-0.1, 0),
                (1500, 1800),
                (-1500, -1800),
                (1800, 1800),
                (-1800, -1800),
                (2000, 1800),
                (-2000, -1800),
                (5 * 3600 - 100, 5 * 3600),
                (-5 * 3600 + 100, -5 * 3600),
            ]
            for time_shift, expected_time_shift in test_data:
                calculated_time_shift = rounded_time_shift(time_shift)
                assert calculated_time_shift == expected_time_shift

    def test_assert_valid_time_shift(self):
        """
        Test time shift sanity checks.
        """
        script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        multisession_factory = scripted_session.factory(script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            # Use private bound method.
            assert_time_shift = host._FTPHost__assert_valid_time_shift
            # Valid time shifts
            test_data = [23 * 3600, -23 * 3600, 3600 + 30, -3600 + 30]
            for time_shift in test_data:
                assert assert_time_shift(time_shift) is None
            # Invalid time shift (exceeds one day)
            with pytest.raises(ftputil.error.TimeShiftError):
                assert_time_shift(25 * 3600)
            with pytest.raises(ftputil.error.TimeShiftError):
                assert_time_shift(-25 * 3600)
            # Invalid time shift (too large deviation from 15-minute units is
            # unacceptable)
            with pytest.raises(ftputil.error.TimeShiftError):
                assert_time_shift(8 * 60)
            with pytest.raises(ftputil.error.TimeShiftError):
                assert_time_shift(-3600 - 8 * 60)

    def test_synchronize_times(self):
        """
        Test time synchronization with server.
        """
        host_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("delete", args=("_ftputil_sync_",)),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call(
                "transfercmd", args=("STOR _ftputil_sync_", None), result=io.BytesIO()
            ),
            Call("voidresp"),
            Call("close"),
        ]
        # Valid time shifts
        test_data = [
            (60 * 60 + 30, 60 * 60),
            (60 * 60 - 100, 60 * 60),
            (30 * 60 + 100, 30 * 60),
            (45 * 60 - 100, 45 * 60),
        ]
        for measured_time_shift, expected_time_shift in test_data:
            # Use a new `BytesIO` object to avoid exception
            # `ValueError: I/O operation on closed file`.
            file_script[4] = Call(
                "transfercmd", result=io.BytesIO(), args=("STOR _ftputil_sync_", None)
            )
            multisession_factory = scripted_session.factory(host_script, file_script)
            with test_base.ftp_host_factory(multisession_factory) as host:
                host.path = self._Path()
                host.path.set_mtime(time.time() + measured_time_shift)
                host.synchronize_times()
                assert host.time_shift() == expected_time_shift
        # Invalid time shifts
        measured_time_shifts = [60 * 60 + 8 * 60, 45 * 60 - 6 * 60]
        for measured_time_shift in measured_time_shifts:
            # Use a new `BytesIO` object to avoid exception
            # `ValueError: I/O operation on closed file`.
            file_script[4] = Call(
                "transfercmd", result=io.BytesIO(), args=("STOR _ftputil_sync_", None)
            )
            multisession_factory = scripted_session.factory(host_script, file_script)
            with test_base.ftp_host_factory(multisession_factory) as host:
                host.path = self._Path()
                host.path.set_mtime(time.time() + measured_time_shift)
                with pytest.raises(ftputil.error.TimeShiftError):
                    host.synchronize_times()

    def test_synchronize_times_for_server_in_east(self):
        """
        Test for timestamp correction (see ticket #55).
        """
        host_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("delete", args=("_ftputil_sync_",)),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call(
                "transfercmd", args=("STOR _ftputil_sync_", None), result=io.BytesIO()
            ),
            Call("voidresp", args=()),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(session_factory=multisession_factory) as host:
            host.path = self._Path()
            # Set this explicitly to emphasize the problem.
            host.set_time_shift(0.0)
            hour = 60 * 60
            # This could be any negative time shift.
            presumed_time_shift = -6 * hour
            # Set `mtime` to simulate a server east of us.
            # In case the `time_shift` value for this host instance is 0.0
            # (as is to be expected before the time shift is determined), the
            # directory parser (more specifically
            # `ftputil.stat.Parser.parse_unix_time`) will return a time which
            # is a year too far in the past. The `synchronize_times` method
            # needs to deal with this and add the year "back". I don't think
            # this is a bug in `parse_unix_time` because the method should work
            # once the time shift is set correctly.
            client_time = datetime.datetime.utcnow().replace(
                tzinfo=datetime.timezone.utc
            )
            presumed_server_time = client_time.replace(
                year=client_time.year - 1
            ) + datetime.timedelta(seconds=presumed_time_shift)
            host.path.set_mtime(presumed_server_time.timestamp())
            host.synchronize_times()
            assert host.time_shift() == presumed_time_shift


class TestUploadAndDownload:
    """
    Test upload and download.
    """

    def test_download(self, tmp_path):
        """
        Test mode download.
        """
        remote_file_name = "dummy_name"
        remote_file_content = b"dummy_content"
        local_target = tmp_path / "test_target"
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call(
                "transfercmd",
                args=("RETR {}".format(remote_file_name), None),
                result=io.BytesIO(remote_file_content),
            ),
            Call("voidresp"),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        # Download
        with test_base.ftp_host_factory(multisession_factory) as host:
            host.download(remote_file_name, str(local_target))
        # Verify expected operations on mock socket as done in `FTPFile.close`.
        # We expect one `gettimeout` and two `settimeout` calls.
        file_session = multisession_factory.scripted_sessions[1]
        file_session.sock.gettimeout.assert_called_once_with()
        assert len(file_session.sock.settimeout.call_args_list) == 2
        assert file_session.sock.settimeout.call_args_list[0] == (
            (ftputil.file.FTPFile._close_timeout,),
            {},
        )
        assert file_session.sock.settimeout.call_args_list[1] == (
            (file_session.sock.gettimeout(),),
            {},
        )
        assert local_target.read_bytes() == remote_file_content

    def test_conditional_upload_without_upload(self, tmp_path):
        """
        If the target file is newer, no upload should happen.
        """
        local_source = tmp_path / "test_source"
        data = binary_data()
        local_source.write_bytes(data)
        dir_result = test_base.dir_line(
            mode_string="-rw-r--r--",
            date_=datetime.date.today() + datetime.timedelta(days=1),
            name="newer",
        )
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_result),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        # Target is newer, so don't upload.
        #
        # This not only tests the return value, but also if a transfer
        # happened. If an upload was tried, our test framework would complain
        # about a missing scripted session for the `FTPFile` host.
        multisession_factory = scripted_session.factory(script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            flag = host.upload_if_newer(str(local_source), "/newer")
        assert flag is False

    def test_conditional_upload_with_upload(self, tmp_path):
        """
        If the target file is older or doesn't exist, the source file should be
        uploaded.
        """
        local_source = tmp_path / "test_source"
        file_content = b"dummy_content"
        local_source.write_bytes(file_content)
        remote_file_name = "dummy_name"
        dir_result = test_base.dir_line(
            mode_string="-rw-r--r--",
            date_=datetime.date.today() - datetime.timedelta(days=1),
            name="older",
        )
        host_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_result),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call(
                "transfercmd",
                args=("STOR older", None),
                result=test_base.MockableBytesIO(),
            ),
            Call("voidresp"),
            Call("close"),
        ]
        # Target is older, so upload.
        multisession_factory = scripted_session.factory(host_script, file_script)
        with unittest.mock.patch("test.test_base.MockableBytesIO.write") as write_mock:
            with test_base.ftp_host_factory(multisession_factory) as host:
                flag = host.upload_if_newer(str(local_source), "/older")
            write_mock.assert_called_with(file_content)
        assert flag is True
        # Target doesn't exist, so upload.
        #  Use correct file name for this test.
        file_script[4] = Call(
            "transfercmd",
            args=("STOR notthere", None),
            result=test_base.MockableBytesIO(),
        )
        multisession_factory = scripted_session.factory(host_script, file_script)
        with unittest.mock.patch("test.test_base.MockableBytesIO.write") as write_mock:
            with test_base.ftp_host_factory(multisession_factory) as host:
                flag = host.upload_if_newer(str(local_source), "/notthere")
            write_mock.assert_called_with(file_content)
        assert flag is True

    def test_conditional_download_without_target(self, tmp_path):
        """
        Test conditional binary mode download when no target file exists.
        """
        local_target = tmp_path / "test_target"
        data = binary_data()
        # Target does not exist, so download.
        #  There isn't a `dir` call to compare the datetimes of the remote and
        #  the target file because the local `exists` call for the local target
        #  returns `False` and the datetime comparison therefore isn't done.
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call("transfercmd", args=("RETR newer", None), result=io.BytesIO(data)),
            Call("voidresp"),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            flag = host.download_if_newer("/newer", str(local_target))
        assert flag is True
        assert local_target.read_bytes() == data

    def test_conditional_download_with_older_target(self, tmp_path):
        """
        Test conditional binary mode download with newer source file.
        """
        local_target = tmp_path / "test_target"
        # Make sure file exists for the timestamp comparison.
        local_target.touch()
        data = binary_data()
        # Target is older, so download.
        #  Use a date in the future. That isn't realistic, but for the purpose
        #  of the test it's an easy way to make sure the source file is newer
        #  than the target file.
        dir_result = test_base.dir_line(
            mode_string="-rw-r--r--",
            date_=datetime.date.today() + datetime.timedelta(days=1),
            name="newer",
        )
        host_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_result),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call("transfercmd", args=("RETR newer", None), result=io.BytesIO(data)),
            Call("voidresp"),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            flag = host.download_if_newer("/newer", str(local_target))
        assert flag is True
        assert local_target.read_bytes() == data

    def test_conditional_download_with_newer_target(self, tmp_path):
        """
        Test conditional binary mode download with older source file.
        """
        local_target = tmp_path / "test_target"
        # Make sure file exists for timestamp comparison.
        local_target.touch()
        data = binary_data()
        # Use date in the past, so the target file is newer and no download
        # happens.
        dir_result = test_base.dir_line(
            mode_string="-rw-r--r--",
            date_=datetime.date.today() - datetime.timedelta(days=1),
            name="newer",
        )
        host_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_result),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call("transfercmd", args=("RETR newer", None), result=io.BytesIO(data)),
            Call("voidresp"),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            flag = host.download_if_newer("/newer", str(local_target))
        assert flag is False


class TestMakedirs:
    def test_exist_ok_false(self):
        """
        If `exist_ok` is `False` or not specified, an existing leaf directory
        should lead to a `PermanentError` with `errno` set to 17.
        """
        # No `exist_ok` specified
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/part1",)),
            Call("cwd", args=("/part1/part2",)),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(script)
        with test_base.ftp_host_factory(session_factory=multisession_factory) as host:
            with pytest.raises(ftputil.error.PermanentError) as exc_info:
                host.makedirs("/part1/part2")
            assert isinstance(exc_info.value, ftputil.error.PermanentError)
            assert exc_info.value.errno == errno.EEXIST
        # `exist_ok` explicitly set to `False`
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/part1",)),
            Call("cwd", args=("/part1/part2",)),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(script)
        with test_base.ftp_host_factory(session_factory=multisession_factory) as host:
            with pytest.raises(ftputil.error.PermanentError) as exc_info:
                host.makedirs("/part1/part2", exist_ok=False)
            assert isinstance(exc_info.value, ftputil.error.PermanentError)
            assert exc_info.value.errno == errno.EEXIST

    def test_exist_ok_true(self):
        """
        If `exist_ok` is `True`, an existing leaf directory should _not_ lead
        to an exception.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/part1",)),
            Call("cwd", args=("/part1/part2",)),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(script)
        with test_base.ftp_host_factory(session_factory=multisession_factory) as host:
            host.makedirs("/part1/part2", exist_ok=True)


class TestAcceptEitherUnicodeOrBytes:
    """
    Test whether certain `FTPHost` methods accept either unicode
    or byte strings for the path(s).
    """

    def test_upload(self):
        """
        Test whether `upload` accepts either unicode or bytes.
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call("transfercmd", args=("STOR target", None), result=io.BytesIO()),
            Call("voidresp"),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        # The source file needs to be present in the current directory.
        with test_base.ftp_host_factory(multisession_factory) as host:
            host.upload("Makefile", "target")
        # Create new `BytesIO` object.
        file_script[4] = Call(
            "transfercmd", args=("STOR target", None), result=io.BytesIO()
        )
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            host.upload("Makefile", as_bytes("target"))

    def test_download(self, tmp_path):
        """
        Test whether `download` accepts either unicode or bytes.
        """
        local_target = tmp_path / "local_target"
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call("transfercmd", args=("RETR source", None), result=io.BytesIO()),
            Call("voidresp"),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        # The source file needs to be present in the current directory.
        with test_base.ftp_host_factory(multisession_factory) as host:
            host.download("source", str(local_target))
        # Create new `BytesIO` object.
        file_script[4] = Call(
            "transfercmd", args=("RETR source", None), result=io.BytesIO()
        )
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            host.download(as_bytes("source"), str(local_target))

    def test_rename(self):
        """
        Test whether `rename` accepts either unicode or bytes.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("rename", args=("/ä", "/ä")),
            Call("close"),
        ]
        # It's possible to mix argument types, as for `os.rename`.
        path_as_str = "/ä"
        path_as_bytes = as_bytes(
            path_as_str, ftputil.path_encoding.FTPLIB_DEFAULT_ENCODING
        )
        paths = [path_as_str, path_as_bytes]
        for source_path, target_path in itertools.product(paths, paths):
            # Uses ftplib default encoding
            session_factory = scripted_session.factory(script)
            with test_base.ftp_host_factory(session_factory) as host:
                host.rename(source_path, target_path)

    def test_listdir(self):
        """
        Test whether `listdir` accepts either unicode or bytes.
        """
        top_level_dir_line = test_base.dir_line(
            mode_string="drwxr-xr-x", date_=datetime.date.today(), name="ä"
        )
        dir_line1 = test_base.dir_line(
            mode_string="-rw-r--r--", date_=datetime.date.today(), name="ö"
        )
        dir_line2 = test_base.dir_line(
            mode_string="-rw-r--r--", date_=datetime.date.today(), name="o"
        )
        dir_result = dir_line1 + "\n" + dir_line2
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=top_level_dir_line),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/ä",)),
            Call("dir", args=("",), result=dir_result),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        # Unicode
        session_factory = scripted_session.factory(script)
        with test_base.ftp_host_factory(session_factory) as host:
            items = host.listdir("ä")
        assert items == ["ö", "o"]
        # Bytes
        session_factory = scripted_session.factory(script)
        with test_base.ftp_host_factory(session_factory) as host:
            items = host.listdir(
                as_bytes("ä", ftputil.path_encoding.FTPLIB_DEFAULT_ENCODING)
            )
            assert items == [
                as_bytes("ö", ftputil.path_encoding.FTPLIB_DEFAULT_ENCODING),
                as_bytes("o"),
            ]

    def test_chmod(self):
        """
        Test whether `chmod` accepts either unicode or bytes.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("SITE CHMOD 0755 ä",)),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        path = "/ä"
        # Unicode
        session_factory = scripted_session.factory(script)
        with test_base.ftp_host_factory(session_factory) as host:
            host.chmod(path, 0o755)
        # Bytes
        session_factory = scripted_session.factory(script)
        with test_base.ftp_host_factory(session_factory) as host:
            host.chmod(
                as_bytes(path, ftputil.path_encoding.FTPLIB_DEFAULT_ENCODING), 0o755
            )

    def _test_method_with_single_path_argument(self, method_name, path, script):
        # Unicode
        session_factory = scripted_session.factory(script)
        with test_base.ftp_host_factory(session_factory) as host:
            method = getattr(host, method_name)
            method(path)
        # Bytes
        session_factory = scripted_session.factory(script)
        with test_base.ftp_host_factory(session_factory) as host:
            method = getattr(host, method_name)
            method(as_bytes(path, ftputil.path_encoding.FTPLIB_DEFAULT_ENCODING))

    def test_chdir(self):
        """
        Test whether `chdir` accepts either unicode or bytes.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/ö",)),
            Call("close"),
        ]
        self._test_method_with_single_path_argument("chdir", "/ö", script)

    def test_mkdir(self):
        """
        Test whether `mkdir` accepts either unicode or bytes.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("mkd", args=("ä",)),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        self._test_method_with_single_path_argument("mkdir", "/ä", script)

    def test_makedirs(self):
        """
        Test whether `makedirs` accepts either unicode or bytes.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            # To deal with ticket #86 (virtual directories), `makedirs` tries
            # to change into each directory and if it exists (changing doesn't
            # raise an exception), doesn't try to create it. That's why you
            # don't see an `mkd` calls here despite originally having a
            # `makedirs` call.
            Call("cwd", args=("/ä",)),
            # If `exist_ok` is `False` (which is the default), the leaf
            # directory to make must not exist. In other words, the `chdir`
            # call is `makedirs` must fail with a permanent error.
            Call("cwd", args=("/ä/ö",), result=ftplib.error_perm),
            Call("cwd", args=("/ä",)),
            Call("cwd", args=("/ä",)),
            Call("mkd", args=("ö",)),
            # From `isdir` call
            Call("cwd", args=("/ä",)),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        self._test_method_with_single_path_argument("makedirs", "/ä/ö", script)

    def test_rmdir(self):
        """
        Test whether `rmdir` accepts either unicode or bytes.
        """
        dir_line = test_base.dir_line(
            mode_string="drwxr-xr-x", date_=datetime.date.today(), name="empty_ä"
        )
        # Since the session script isn't at all obvious, I checked it with a
        # debugger and added comments on some of the calls that happen during
        # the `rmdir` call.
        #
        # `_robust_ftp_command` descends one directory at a time (see ticket
        # #11) and restores the original directory in the end, which results in
        # at least four calls on the FTP session object (`cwd`, `cwd`, actual
        # method, `cwd`). It would be great if all the roundtrips to the server
        # could be reduced.
        script = [
            # `FTPHost` initialization
            Call("__init__"),
            Call("pwd", result="/"),
            # `host.rmdir("/empty_ä")`
            #  `host.listdir("/empty_ä")`
            #   `host._stat._listdir("/empty_ä")`
            #    `host._stat.__call_with_parser_retry("/empty_ä")`
            #     `host._stat._real_listdir("/empty_ä")`
            #      `host.path.isdir("/empty_ä")`
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_line),
            Call("cwd", args=("/",)),
            #      `host.path.isdir` end
            #      `host._stat._stat_results_from_dir("/empty_ä")`
            Call("cwd", args=("/",)),
            Call("cwd", args=("/empty_ä",)),
            Call("dir", args=("",), result=""),
            Call("cwd", args=("/",)),
            #      `host._stat._stat_results_from_dir("/empty_ä")` end
            #  `host._session.rmd` in `host._robust_ftp_command`
            #   `host._check_inaccessible_login_directory()`
            Call("cwd", args=("/",)),
            #   `host.chdir(head)` ("/")
            Call("cwd", args=("/",)),
            #   `host.rmd(tail)` ("empty_ä")
            Call("rmd", args=("empty_ä",)),
            #   `host.chdir(old_dir)` ("/")
            Call("cwd", args=("/",)),
            #
            Call("close"),
        ]
        empty_directory_as_required_by_rmdir = "/empty_ä"
        self._test_method_with_single_path_argument(
            "rmdir", empty_directory_as_required_by_rmdir, script
        )

    def test_remove(self):
        """
        Test whether `remove` accepts either unicode or bytes.
        """
        dir_line = test_base.dir_line(
            mode_string="-rw-r--r--", date_=datetime.date.today(), name="ö"
        )
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_line),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("delete", args=("ö",)),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        self._test_method_with_single_path_argument("remove", "/ö", script)

    def test_rmtree(self):
        """
        Test whether `rmtree` accepts either unicode or bytes.
        """
        dir_line = test_base.dir_line(
            mode_string="drwxr-xr-x", date_=datetime.date.today(), name="empty_ä"
        )
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            # Recursive `listdir`
            #  Check parent (root) directory.
            Call("dir", args=("",), result=dir_line),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/empty_ä",)),
            #  Child directory (inside `empty_ä`)
            Call("dir", args=("",), result=""),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/empty_ä",)),
            # Recursive `rmdir` (repeated `cwd` calls because of
            # `_robust_ftp_command`)
            Call("dir", result="", args=("",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("rmd", args=("empty_ä",)),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        empty_directory_as_required_by_rmtree = "/empty_ä"
        self._test_method_with_single_path_argument(
            "rmtree", empty_directory_as_required_by_rmtree, script
        )

    def test_lstat(self):
        """
        Test whether `lstat` accepts either unicode or bytes.
        """
        dir_line = test_base.dir_line(
            mode_string="-rw-r--r--", date_=datetime.date.today(), name="ä"
        )
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_line),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        self._test_method_with_single_path_argument("lstat", "/ä", script)

    def test_stat(self):
        """
        Test whether `stat` accepts either unicode or bytes.
        """
        dir_line = test_base.dir_line(
            mode_string="-rw-r--r--", date_=datetime.date.today(), name="ä"
        )
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_line),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        self._test_method_with_single_path_argument("stat", "/ä", script)

    def test_walk(self):
        """
        Test whether `walk` accepts either unicode or bytes.
        """
        dir_line = test_base.dir_line(
            mode_string="-rw-r--r--", date_=datetime.date.today(), name="ä"
        )
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_line),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        # We're not interested in the return value of `walk`.
        # Unicode
        session_factory = scripted_session.factory(script)
        with test_base.ftp_host_factory(session_factory) as host:
            result = list(host.walk("/ä"))
        # Bytes
        session_factory = scripted_session.factory(script)
        with test_base.ftp_host_factory(session_factory) as host:
            result = list(
                host.walk(as_bytes("/ä", ftputil.path_encoding.FTPLIB_DEFAULT_ENCODING))
            )


class TestFailingPickling:
    def test_failing_pickling(self):
        """
        Test if pickling (intentionally) isn't supported.
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call("transfercmd", args=("RETR test", None), result=io.BytesIO()),
            Call("voidresp", args=()),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            with pytest.raises(TypeError):
                pickle.dumps(host)
            with host.open("/test") as file_obj:
                with pytest.raises(TypeError):
                    pickle.dumps(file_obj)



# Copyright (C) 2018-2021, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

import sys
import unittest.mock

import ftputil.path_encoding


__all__ = ["Call", "factory"]


class Call:
    def __init__(self, method_name, *, args=None, kwargs=None, result=None):
        self.method_name = method_name
        self.result = result
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        return (
            "{0.__class__.__name__}("
            "method_name={0.method_name!r}, "
            "result={0.result!r}, "
            "args={0.args!r}, "
            "kwargs={0.kwargs!r})".format(self)
        )

    def check_call(self, method_name, args=None, kwargs=None):
        # TODO: Mention printing in the docstring.
        # TODO: Describe how the comparison is made.
        """
        Check the method name, args and kwargs from this `Call` object against
        the method name, args and kwargs from the system under test.

        Raise an `AssertionError` if there's a mismatch.
        """
        print(
            "  Call from session script:    {} | {!r} | {!r}".format(
                self.method_name, self.args, self.kwargs
            )
        )
        print(
            "  Call from system under test: {} | {!r} | {!r}".format(
                method_name, args, kwargs
            )
        )

        def compare(value_name, script_value, sut_value):
            if script_value is not None:
                try:
                    assert script_value == sut_value
                except AssertionError:
                    print(
                        "  Mismatch for `{}`: {!r} != {!r}".format(
                            value_name, script_value, sut_value
                        )
                    )
                    raise

        compare("method_name", self.method_name, method_name)
        compare("args", self.args, args)
        compare("kwargs", self.kwargs, kwargs)

    @staticmethod
    def _is_exception_class(obj):
        """
        Return `True` if `obj` is an exception class, else `False`.
        """
        try:
            return issubclass(obj, Exception)
        except TypeError:
            # TypeError: issubclass() arg 1 must be a class
            return False

    def __call__(self):
        """
        Simulate call, returning the result or raising the exception.
        """
        if isinstance(self.result, Exception) or self._is_exception_class(self.result):
            raise self.result
        else:
            return self.result


class ScriptedSession:
    """
    "Scripted" `ftplib.FTP`-like class for testing.

    To avoid actual input/output over sockets or files, specify the values that
    should be returned by the class's methods.

    The class is instantiated with a `script` argument. This is a list of
    `Call` objects where each object specifies the name of the `ftplib.FTP`
    method that is expected to be called and what the method should return. If
    the value is an exception, it will be raised, not returned.

    In case the method returns a socket (like `transfercmd`), the return value
    to be specified in the `Call` instance is the content of the underlying
    socket file.

    The advantage of the approach of this class over the use of
    `unittest.mock.Mock` objects is that the sequence of calls is clearly
    visible. With `Mock` objects, the developer must keep in mind all the calls
    when specifying return values or side effects for the mock methods.
    """

    # Class-level counter to enumerate `ScriptedSession`s. This makes it
    # possible to make the output even more compact. Additionally, it's easier
    # to distinguish numbers like 1, 2, etc. than hexadecimal ids.
    _session_count = 0

    encoding = ftputil.path_encoding.FTPLIB_DEFAULT_ENCODING

    @classmethod
    def reset_session_count(cls):
        cls._session_count = 0

    def __init__(self, script):
        self.script = script
        # `File.close` accesses the session `sock` object to set and reset the
        # timeout. `sock` itself is never _called_ though, so it doesn't make
        # sense to create a `sock` _call_.
        self.sock = unittest.mock.Mock(name="socket_attribute")
        # Index into `script`, the list of `Call` objects
        self._call_index = 0
        self.__class__._session_count += 1
        self._session_count = self.__class__._session_count
        # Always expect an entry for the constructor.
        init_call = self._next_script_call("__init__")
        # The constructor isn't supposed to return anything. The only reason to
        # call it here is to raise an exception if that was specified in the
        # `script`.
        init_call()

    def __str__(self):
        return "{} {}".format(self.__class__.__name__, self._session_count)

    def _next_script_call(self, requested_attribute):
        """
        Return next `Call` object.
        """
        print(self, "in `_next_script_call`")
        try:
            call = self.script[self._call_index]
        except IndexError:
            print("  *** Ran out of `Call` objects for this session {!r}".format(self))
            print("  Requested attribute was {!r}".format(requested_attribute))
            raise
        self._call_index += 1
        print(self, f"next call: {call!r}")
        return call

    def __getattr__(self, attribute_name):
        script_call = self._next_script_call(attribute_name)

        def dummy_method(*args, **kwargs):
            print(self, "in `__getattr__`")
            script_call.check_call(attribute_name, args, kwargs)
            return script_call()

        return dummy_method

    # ----------------------------------------------------------------------
    # `ftplib.FTP` methods that shouldn't be executed with the default
    # processing in `__getattr__`

    def dir(self, path, callback):
        """
        Call the `callback` for each line in the multiline string
        `call.result`.
        """
        script_call = self._next_script_call("dir")
        # Check only the path. This requires that the corresponding `Call`
        # object also solely specifies the path as `args`.
        script_call.check_call("dir", (path,), None)
        # Give `dir` the chance to raise an exception if one was specified in
        # the `Call`'s `result` argument.
        call_result = script_call()
        for line in call_result.splitlines():
            callback(line)

    def ntransfercmd(self, cmd, rest=None):
        """
        Simulate the `ftplib.FTP.ntransfercmd` call.

        `ntransfercmd` returns a tuple of a socket and a size argument. The
        `result` value given when constructing an `ntransfercmd` call specifies
        an `io.TextIO` or `io.BytesIO` value to be used as the
        `Socket.makefile` result.
        """
        script_call = self._next_script_call("ntransfercmd")
        script_call.check_call("ntransfercmd", (cmd, rest), None)
        # Give `ntransfercmd` the chance to raise an exception if one was
        # specified in the `Call`'s `result` argument.
        call_result = script_call()
        mock_socket = unittest.mock.Mock(name="socket")
        mock_socket.makefile.return_value = call_result
        # Return `None` for size. The docstring of `ftplib.FTP.ntransfercmd`
        # says that's a possibility.
        # TODO: Use a sensible `size` value later if it turns out we need it.
        return mock_socket, None

    def transfercmd(self, cmd, rest=None):
        """
        Simulate the `ftplib.FTP.transfercmd` call.

        `transfercmd` returns a socket. The `result` value given when
        constructing an `transfercmd` call specifies an `io.TextIO` or
        `io.BytesIO` value to be used as the `Socket.makefile` result.
        """
        script_call = self._next_script_call("transfercmd")
        script_call.check_call("transfercmd", (cmd, rest), None)
        # Give `transfercmd` the chance to raise an exception if one was
        # specified in the `Call`'s `result` argument.
        call_result = script_call()
        mock_socket = unittest.mock.Mock(name="socket")
        mock_socket.makefile.return_value = call_result
        return mock_socket


class MultisessionFactory:
    """
    Return a session factory using the scripted data from the given "scripts"
    for each consecutive call ("creation") of a factory.

    Example:

      host = ftputil.FTPHost(host, user, password,
                             session_factory=scripted_session.factory(script1, script2))

    When the `session_factory` is "instantiated" for the first time by
    `FTPHost._make_session`, the factory object will use the behavior described
    by the script `script1`. When the `session_factory` is "instantiated" a
    second time, the factory object will use the behavior described by the
    script `script2`.
    """

    def __init__(self, *scripts):
        ScriptedSession.reset_session_count()
        self._scripts = iter(scripts)
        self.scripted_sessions = []

    def __call__(self, host, user, password):
        """
        Call the factory.

        This is equivalent to the constructor of the session (e. g.
        `ftplib.FTP` in a real application).
        """
        script = next(self._scripts)
        scripted_session = ScriptedSession(script)
        self.scripted_sessions.append(scripted_session)
        return scripted_session


factory = MultisessionFactory



# Copyright (C) 2009-2018, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

import os
import socket
import subprocess

import pytest

import ftputil

import test


def email_address():
    """
    Return the email address used to identify the client to an FTP server.

    If the hostname is "warpy", use my (Stefan's) email address, else try to
    use the content of the `$EMAIL` environment variable. If that doesn't
    exist, use a dummy address.
    """
    hostname = socket.gethostname()
    if hostname == "warpy":
        email = "sschwarzer@sschwarzer.net"
    else:
        dummy_address = "anonymous@example.com"
        email = os.environ.get("EMAIL", dummy_address)
        if not email:
            # Environment variable exists but content is an empty string
            email = dummy_address
    return email


EMAIL = email_address()


def ftp_client_listing(server, directory):
    """
    Log into the FTP server `server` using the command line client, then change
    to the `directory` and retrieve a listing with "dir".

    Return the list of items found as an `os.listdir` would return it.
    """
    # The `-n` option prevents an auto-login.
    ftp_popen = subprocess.Popen(
        ["ftp", "-n", server],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        universal_newlines=True,
    )
    commands = ["user anonymous {}".format(EMAIL), "dir", "bye"]
    if directory:
        # Change to this directory before calling "dir".
        commands.insert(1, "cd {}".format(directory))
    input_ = "\n".join(commands)
    stdout, unused_stderr = ftp_popen.communicate(input_)
    # Collect the directory/file names from the listing's text
    names = []
    for line in stdout.strip().split("\n"):
        if line.startswith("total ") or line.startswith("Trying "):
            continue
        parts = line.split()
        if parts[-2] == "->":
            # Most likely a link
            name = parts[-3]
        else:
            name = parts[-1]
        names.append(name)
    # Remove entries for current and parent directory since they aren't
    # included in the result of `FTPHost.listdir` either.
    names = [name for name in names if name not in (".", "..")]
    return names


class TestPublicServers:
    """
    Get directory listings from various public FTP servers with a command line
    client and ftputil and compare both.

    An important aspect is to test different "spellings" of the same directory.
    For example, to list the root directory which is usually set after login,
    use "" (nothing), ".", "/", "/.", "./.", "././", "..", "../.", "../.." etc.

    The command line client `ftp` has to be in the path.
    """

    # Implementation note:
    #
    # I (Stefan) implement the code so it works with Ubuntu's client. Other
    # clients may work or not. If you have problems testing some other client,
    # please send me a (small) patch. Keep in mind that I don't plan supporting
    # as many FTP obscure commandline clients as servers. ;-)

    # List of pairs with server name and a directory "guaranteed to exist"
    # under the login directory which is assumed to be the root directory.
    servers = [  # Posix format
        ("ftp.de.debian.org", "debian"),
        # ("ftp.gnome.org", "pub"),
        ("ftp.heanet.ie", "pub"),
        ("ftp.heise.de", "pub"),
        # DOS/Microsoft format
        # Do you know any FTP servers that use Microsoft format?
        # `ftp.microsoft.com` doesn't seem to be reachable anymore.
    ]

    # This data structure contains the initial directories "." and "DIR" (which
    # will be replaced by a valid directory name for each server). The list
    # after the initial directory contains paths that will be queried after
    # changing into the initial directory. All items in these lists are
    # actually supposed to yield the same directory contents.
    paths_table = [
        (
            ".",
            [
                ".",
                "/",
                "/.",
                "./.",
                "././",
                "..",
                "../.",
                "../..",
                "DIR/..",
                "/DIR/../.",
                "/DIR/../..",
            ],
        ),
        ("DIR", [".", "/DIR", "/DIR/", "../DIR", "../../DIR"]),
    ]

    def inner_test_server(self, server, initial_directory, paths):
        """
        Test one server for one initial directory.

        Connect to the server `server`; if the string argument
        `initial_directory` has a true value, change to this directory. Then
        iterate over all strings in the sequence `paths`, comparing the results
        of a listdir call with the listing from the command line client.
        """
        canonical_names = ftp_client_listing(server, initial_directory)
        host = ftputil.FTPHost(server, "anonymous", EMAIL)
        try:
            host.chdir(initial_directory)
            for path in paths:
                path = path.replace("DIR", initial_directory)
                # Make sure that we don't recycle directory entries, i. e.
                # really repeatedly retrieve the directory contents (shouldn't
                # happen anyway with the current implementation).
                host.stat_cache.clear()
                names = host.listdir(path)
                # Filter out "hidden" names since the FTP command line client
                # won't include them in its listing either.
                names = [
                    name
                    for name in names
                    if not (
                        name.startswith(".")
                        or
                        # The login directory of `ftp.microsoft.com` contains
                        # this "hidden" entry that ftputil finds but not the
                        # FTP command line client.
                        name == "mscomtest"
                    )
                ]
                failure_message = "For server {}, directory {}: {} != {}".format(
                    server, initial_directory, names, canonical_names
                )
                assert names == canonical_names, failure_message
        finally:
            host.close()

    @pytest.mark.slow_test
    def test_servers(self):
        """
        Test all servers in `self.servers`.

        For each server, get the listings for the login directory and one other
        directory which is known to exist. Use different "spellings" to
        retrieve each list via ftputil and compare with the results gotten with
        the command line client.
        """
        for server, actual_initial_directory in self.servers:
            print("=== server:", server)
            for initial_directory, paths in self.paths_table:
                initial_directory = initial_directory.replace(
                    "DIR", actual_initial_directory
                )
                print(server, initial_directory)
                self.inner_test_server(server, initial_directory, paths)



# Copyright (C) 2002-2018, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

"""
ftputil - high-level FTP client library

FTPHost objects
    This class resembles the `os` module's interface to ordinary file
    systems. In addition, it provides a method `file` which will
    return file-objects corresponding to remote files.

    # Example session
    with ftputil.FTPHost("ftp.domain.com", "me", "secret") as host:
        print(host.getcwd())  # e. g. "/home/me"
        host.mkdir("newdir")
        host.chdir("newdir")
        with host.open("sourcefile", "r") as source:
            with host.open("targetfile", "w") as target:
                host.copyfileobj(source, target)
        host.remove("targetfile")
        host.chdir(host.pardir)
        host.rmdir("newdir")

    There are also shortcuts for uploads and downloads:

    host.upload(local_file, remote_file)
    host.download(remote_file, local_file)

    Both accept an additional mode parameter. If it is "b", the
    transfer mode will be for binary files.

    For even more functionality refer to the documentation in
    `ftputil.txt` or `ftputil.html`.

FTPFile objects
    `FTPFile` objects are constructed via the `file` method (`open`
    is an alias) of `FTPHost` objects. `FTPFile` objects support the
    usual file operations for non-seekable files (`read`, `readline`,
    `readlines`, `write`, `writelines`, `close`).

Note: ftputil currently is not threadsafe. More specifically, you can
      use different `FTPHost` objects in different threads but not
      a single `FTPHost` object in different threads.
"""

from ftputil.host import FTPHost
from ftputil.version import __version__


# Apart from `ftputil.error` and `ftputil.stat`, this is the whole
# public API of `ftputil`.
__all__ = ["FTPHost", "__version__"]



# Copyright (C) 2002-2019, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

import ftplib
import io
import unittest

import pytest

import ftputil.error

from test import scripted_session
from test import test_base


Call = scripted_session.Call


# Test data for use in text/binary read/write tests.
TEXT_TEST_DATA = " söme line\r\nänother line\n almost done\n"
BINARY_TEST_DATA = TEXT_TEST_DATA.encode("UTF-8")


class TestFileOperations:
    """Test operations with file-like objects."""

    def test_inaccessible_dir(self):
        """
        Test whether opening a file at an invalid location fails.
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call(
                "transfercmd",
                args=("STOR inaccessible", None),
                result=ftplib.error_perm,
            ),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            with pytest.raises(ftputil.error.FTPIOError):
                host.open("/inaccessible", "w")

    def test_caching_of_children(self):
        """
        Test whether `FTPFile` cache of `FTPHost` object works.
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file1_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call("transfercmd", args=("STOR path1", None), result=io.StringIO("")),
            Call("voidresp"),
            # Open a file again while reusing the child object and with it its
            # `_session` attribute (the `ftplib.FTP` object).
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call("transfercmd", args=("STOR path1", None), result=io.StringIO("")),
            Call("voidresp"),
            Call("close"),
        ]
        file2_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call("transfercmd", result=io.StringIO(""), args=("STOR path2", None)),
            Call("voidresp"),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(
            host_script, file1_script, file2_script
        )
        with test_base.ftp_host_factory(multisession_factory) as host:
            assert len(host._children) == 0
            path1 = "path1"
            path2 = "path2"
            # Open one file and inspect cache of available children.
            file1 = host.open(path1, "w")
            child1 = host._children[0]
            assert len(host._children) == 1
            assert not child1._file.closed
            # Open another file.
            file2 = host.open(path2, "w")
            child2 = host._children[1]
            assert len(host._children) == 2
            assert not child2._file.closed
            # Close first file.
            file1.close()
            assert len(host._children) == 2
            assert child1._file.closed
            assert not child2._file.closed
            # Re-open first child's file.
            file1 = host.open(path1, "w")
            child1_1 = file1._host
            # Check if it's reused.
            assert child1 is child1_1
            assert not child1._file.closed
            assert not child2._file.closed
            # Close second file.
            file2.close()
            assert child2._file.closed

    def test_write_to_directory(self):
        """
        Test whether attempting to write to a directory fails.
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call(
                "transfercmd",
                args=("STOR some_directory", None),
                result=ftplib.error_perm,
            ),
            # Because of the exception, `voidresp` isn't called.
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            with pytest.raises(ftputil.error.FTPIOError):
                host.open("/some_directory", "w")

    def test_mode_not_given(self):
        """
        If the mode isn't given, a file should be opened in text read mode.
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call(
                "transfercmd",
                args=("RETR some_file", None),
                result=io.StringIO(TEXT_TEST_DATA),
            ),
            Call("voidresp"),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            with host.open("some_file") as fobj:
                data = fobj.read()
            assert data == TEXT_TEST_DATA

    def test_mode_is_none(self):
        """
        If the mode is passed as `None`, a `TypeError` should be raised.

        (This is Python's behavior for local files.)
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            with pytest.raises(TypeError):
                with host.open("some_file", None) as fobj:
                    pass

    def test_binary_read(self):
        """
        Read data from a binary file.
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call(
                "transfercmd",
                args=("RETR some_file", None),
                result=io.BytesIO(BINARY_TEST_DATA),
            ),
            Call("voidresp"),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            with host.open("some_file", "rb") as fobj:
                data = fobj.read()
            assert data == BINARY_TEST_DATA

    def test_binary_write(self):
        """
        Write binary data with `write`.
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call(
                "transfercmd",
                args=("STOR some_file", None),
                result=test_base.MockableBytesIO(b""),
            ),
            Call("voidresp"),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with unittest.mock.patch("test.test_base.MockableBytesIO.write") as write_mock:
            with test_base.ftp_host_factory(multisession_factory) as host:
                with host.open("some_file", "wb") as output:
                    output.write(BINARY_TEST_DATA)
            write_mock.assert_called_with(BINARY_TEST_DATA)

    def test_text_read(self):
        """
        Read text with plain `read`.
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call(
                "transfercmd",
                args=("RETR dummy", None),
                # Use the same `newline` as `host.open` call. The default for
                # `StringIO`'s `newline` differs from the default for `open`
                # and `socket.makefile`.
                result=io.StringIO(TEXT_TEST_DATA, newline=None),
            ),
            Call("voidresp"),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            with host.open("dummy", "r") as input_:
                data = input_.read(0)
                assert data == ""
                data = input_.read(3)
                assert data == " sö"
                # Specifically check the behavior around the line ending
                # character.
                data = input_.read(7)
                assert data == "me line"
                data = input_.read(1)
                assert data == "\n"
                data = input_.read()
                assert data == "änother line\n almost done\n"
                data = input_.read()
                assert data == ""

    def test_text_write(self):
        """
        Write text with `write`.
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call(
                "transfercmd",
                args=("STOR dummy", None),
                # Use the same `newline` as `host.open` call. The default for
                # `StringIO`'s `newline` differs from the default for `open`
                # and `socket.makefile`.
                result=test_base.MockableStringIO(newline="\r\n"),
            ),
            Call("voidresp"),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with unittest.mock.patch("test.test_base.MockableStringIO.write") as write_mock:
            with test_base.ftp_host_factory(multisession_factory) as host:
                with host.open("dummy", "w", newline="\r\n") as output:
                    output.write(TEXT_TEST_DATA)
        write_mock.assert_called_with(TEXT_TEST_DATA)

    def test_text_writelines(self):
        """
        Write text with `writelines`.
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call(
                "transfercmd",
                args=("STOR dummy", None),
                # Use the same `newline` as `host.open` call. The default for
                # `StringIO`'s `newline` differs from the default for `open`
                # and `socket.makefile`.
                result=test_base.MockableStringIO(newline="\r\n"),
            ),
            Call("voidresp"),
            Call("close"),
        ]
        data = [" söme line\r\n", "änother line\n", " almost done\n"]
        print("=== data:", data)
        backup_data = data[:]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with unittest.mock.patch(
            "test.test_base.MockableStringIO.writelines"
        ) as write_mock:
            with test_base.ftp_host_factory(multisession_factory) as host:
                with host.open("dummy", "w", newline="\r\n") as output:
                    output.writelines(data)
        write_mock.assert_called_with(data)
        # Ensure that the original data was not modified.
        assert data == backup_data

    def test_binary_readline(self):
        """
        Read binary data with `readline`.
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call(
                "transfercmd",
                args=("RETR dummy", None),
                result=io.BytesIO(BINARY_TEST_DATA),
            ),
            Call("voidresp"),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            with host.open("dummy", "rb") as input_:
                data = input_.readline(3)
                assert data == b" s\xc3"
                data = input_.readline(9)
                assert data == b"\xb6me line\r"
                # 30 = at most 30 bytes
                data = input_.readline(30)
                assert data == b"\n"
                data = input_.readline()
                assert data == b"\xc3\xa4nother line\n"
                data = input_.readline()
                assert data == b" almost done\n"
                data = input_.readline()
                assert data == b""

    def test_text_readline(self):
        """
        Read text with `readline`.
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call(
                "transfercmd",
                args=("RETR dummy", None),
                # Use the same `newline` as `host.open` call. The default for
                # `StringIO`'s `newline` differs from the default for `open`
                # and `socket.makefile`.
                result=io.StringIO(TEXT_TEST_DATA, newline=None),
            ),
            Call("voidresp"),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            with host.open("dummy", "r") as input_:
                data = input_.readline(3)
                assert data == " sö"
                data = input_.readline(7)
                assert data == "me line"
                data = input_.readline(10)
                # Does implicit newline conversion.
                assert data == "\n"
                # 30 = at most 30 bytes
                data = input_.readline(30)
                assert data == "änother line\n"
                data = input_.readline()
                assert data == " almost done\n"
                data = input_.readline()
                assert data == ""

    def test_text_readlines(self):
        """
        Read text with `readlines`.
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call(
                "transfercmd",
                args=("RETR dummy", None),
                # Use the same `newline` as `host.open` call. The default for
                # `StringIO`'s `newline` differs from the default for `open`
                # and `socket.makefile`.
                result=io.StringIO(TEXT_TEST_DATA, newline=None),
            ),
            Call("voidresp"),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            with host.open("dummy", "r") as input_:
                data = input_.read(3)
                assert data == " sö"
                data = input_.readlines()
                assert data == ["me line\n", "änother line\n", " almost done\n"]
                input_.close()

    def test_binary_iterator(self):
        """
        Test iterator interface of `FTPFile` objects for binary files.
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call(
                "transfercmd",
                args=("RETR dummy", None),
                result=io.BytesIO(BINARY_TEST_DATA),
            ),
            Call("voidresp"),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            with host.open("dummy", "rb") as input_:
                input_iterator = iter(input_)
                assert next(input_iterator) == b" s\xc3\xb6me line\r\n"
                # The last two lines don't have a `\r`.
                assert next(input_iterator) == b"\xc3\xa4nother line\n"
                assert next(input_iterator) == b" almost done\n"
                with pytest.raises(StopIteration):
                    input_iterator.__next__()

    def test_text_iterator(self):
        """
        Test iterator interface of `FTPFile` objects for text files.
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call(
                "transfercmd",
                args=("RETR dummy", None),
                # Use the same `newline` as `host.open` call. The default for
                # `StringIO`'s `newline` differs from the default for `open`
                # and `socket.makefile`.
                result=io.StringIO(TEXT_TEST_DATA, newline=None),
            ),
            Call("voidresp"),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            with host.open("dummy", "r") as input_:
                input_iterator = iter(input_)
                assert next(input_iterator) == " söme line\n"
                assert next(input_iterator) == "änother line\n"
                assert next(input_iterator) == " almost done\n"
                with pytest.raises(StopIteration):
                    input_iterator.__next__()

    def test_read_unknown_file(self):
        """
        Test whether reading a file which isn't there fails.
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call("transfercmd", args=("RETR notthere", None), result=ftplib.error_perm),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            with pytest.raises(ftputil.error.FTPIOError):
                with host.open("notthere", "r") as input_:
                    pass


class TestAvailableChild:
    def _failing_pwd(self, exception_class):
        """
        Return a function that will be used instead of the `session.pwd` and
        will raise the exception `exception_to_raise`.
        """

        def new_pwd():
            raise exception_class("")

        return new_pwd

    def _test_with_pwd_error(self, exception_class):
        """
        Test if reusing a child session fails because of
        `child_host._session.pwd` raising an exception of type
        `exception_class`.
        """
        host_script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        first_file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call("transfercmd", args=("RETR dummy1", None), result=io.StringIO("")),
            Call("voidresp"),
            # This `pwd` is executed from `FTPHost._available_child`.
            Call("pwd", result=exception_class),
            Call("close"),
        ]
        second_file_script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("voidcmd", args=("TYPE I",)),
            Call("transfercmd", args=("RETR dummy2", None), result=io.StringIO("")),
            Call("voidresp"),
            Call("close"),
        ]
        multisession_factory = scripted_session.factory(
            host_script, first_file_script, second_file_script
        )
        with test_base.ftp_host_factory(multisession_factory) as host:
            # Implicitly create a child session.
            with host.open("/dummy1") as _:
                pass
            assert len(host._children) == 1
            # Try to create a new file. Since `pwd` in
            # `FTPHost._available_child` raises an exception, a new child
            # session should be created.
            with host.open("/dummy2") as _:
                pass
            assert len(host._children) == 2

    def test_pwd_with_error_temp(self):
        """
        Test if an `error_temp` in `_session.pwd` skips the child session.
        """
        self._test_with_pwd_error(ftplib.error_temp)

    def test_pwd_with_error_reply(self):
        """
        Test if an `error_reply` in `_session.pwd` skips the child session.
        """
        self._test_with_pwd_error(ftplib.error_reply)

    def test_pwd_with_OSError(self):
        """
        Test if an `OSError` in `_session.pwd` skips the child session.
        """
        self._test_with_pwd_error(OSError)

    def test_pwd_with_EOFError(self):
        """
        Test if an `EOFError` in `_session.pwd` skips the child session.
        """
        self._test_with_pwd_error(EOFError)
